from .image_creator import ImageCreator


class AptImageCreator(ImageCreator):
    def get_package_string(self, packages):
        # TODO: validation here to ensure shell safety
        apt_string = " ".join([package["recipe"] for package in packages])
        return "apt-get update && apt-get install -y {} && rm -rf /var/lib/apt/lists/*".format(
            apt_string
        )



import logging
import tarfile
import io
from abc import ABC, abstractmethod

from envy.lib.config import ENVY_CONFIG


class ImageCreator(ABC):
    """ Creates a Docker image, given a set of packages.
        This is an abstract class - you may want an AptImageCreator for now.
        See also: ImageFinder
        Args:
            docker (Client): A Docker client
    """

    docker = None

    def __init__(self, docker):
        self.docker = docker

    def build_image_tag(self):
        return "envy-{}".format(ENVY_CONFIG.get_image_hash())

    def create_image(self, packages: [str]):
        """ Create a docker image from a set of packages.
            NOTE: currently not validating packages, or respecting versions

        Argsuments:
            packages (list<string>): A list of packages to install

        Returns:
            string: the Docker image ID
        """
        dockerfile = self.build_dockerfile(packages)
        tar_bytes = self.build_tarball_bytes(dockerfile)
        image, logs = self.docker.images.build(
            custom_context=True, tag=self.build_image_tag(), rm=True, fileobj=tar_bytes
        )
        logging.info(logs)
        return image.id

    def build_tarball_bytes(self, dockerfile):
        """ Build a tarball BytesIO with the Dockerfile
            NOTE: no validation is performed for malicious content!
            Args:
                dockerfile (string): the Dockerfile you want to include
            Returns:
                BytesIO: A bytes IO object that represents an uncompressed tarball, ready to be passed to Docker.
        """
        tar_bytes = io.BytesIO()
        tar_archive = tarfile.open(fileobj=tar_bytes, mode="x")

        # Write the Dockerfile
        docker_data = dockerfile.encode("utf8")
        info = tarfile.TarInfo(name="Dockerfile")
        info.size = len(docker_data)
        tar_archive.addfile(tarinfo=info, fileobj=io.BytesIO(docker_data))

        # Finish the tar archive
        tar_archive.close()
        # Restore the byte buffer
        tar_bytes.seek(0)
        return tar_bytes

    def build_dockerfile(self, packages):
        """ Create a Dockerfile that will result in a reasonable number of layers.
            NOTE: no validation performed on package names
            Args:
                packages (list<string>): The packages you want to install
            Returns:
                string: the string representation of the Dockerfile
        """
        d_file = f"FROM {ENVY_CONFIG.get_base_image()}\n"
        d_file += f"RUN {self.get_package_string(packages)}\n"
        return d_file

    @abstractmethod
    def get_package_string(self, packages):
        pass



from envy.lib.config import ENVY_CONFIG
from envy.lib.state import ENVY_STATE

from .trigger import Trigger


class TriggerPerContainer(Trigger):
    """ Triggers once per container
    """

    def should_trigger(self) -> bool:
        return ENVY_CONFIG.get_container_hash() != ENVY_STATE.get_container_hash()

    def persist_trigger(self):
        pass



from collections import OrderedDict

from envy.lib.docker_manager import ContainerManager
from envy.lib.config import ENVY_CONFIG

import envy.lib.triggers as triggers

from .script_setup_step import ScriptSetupStep
from .remote_setup_step import RemoteSetupStep


class Builder:
    """ Runs triggered build steps on a container
    """

    def __init__(self, container: ContainerManager):
        self.container = container
        self.steps = OrderedDict()

    def build(self):
        # Create system packages step
        self.__create_system_packages_step()

        # Create steps
        self.__create_steps()

        # Run triggered steps
        self.__run_triggered()

        # Persist triggers
        self.__persist_triggers()

    def __create_system_packages_step(self):
        pass

    def __create_steps(self):
        for m in ENVY_CONFIG.get_setup_steps():
            # Create step
            name = m["name"]
            if m["type"] == "script":
                step = ScriptSetupStep(name, self.container, m["run"])
            elif m["type"] == "remote":
                step = RemoteSetupStep(name, self.container, m["url"])

            # Create and register triggers
            if m["triggers"] == "always":
                trigger = triggers.TriggerAlways()
            else:
                trigger_list = []
                for t in m["triggers"]["system-packages"]:
                    # TODO: swap this out once system packages aren't implemented in the image
                    trigger_list.append(triggers.TriggerPerContainer())
                    # trigger_list.append(triggers.TriggerSystemPackage(t))
                for t in m["triggers"]["files"]:
                    trigger_list.append(triggers.TriggerWatchfile(t))
                for t in m["triggers"]["steps"]:
                    trigger_list.append(triggers.TriggerStep(self.steps[t]))

                if trigger_list:
                    trigger = triggers.TriggerGroup(trigger_list)
                else:
                    trigger = triggers.TriggerPerContainer()

            step.set_trigger(trigger)

            # Add step to dict
            self.steps[name] = step

    def __run_triggered(self):
        for step in self.steps.values():
            if step.should_trigger():
                print("Running build step {}".format(step.name))
                step.run()
            else:
                print("Skipping build step {}".format(step.name))

    def __persist_triggers(self):
        for step in self.steps.values():
            if step.has_built():
                step.persist_trigger()



from docker import DockerClient
from docker.models.images import Image

from envy.lib.config import ENVY_CONFIG

from .image_creators import AptImageCreator


class ImageManager:
    ### Static Image Creation ###
    @staticmethod
    def create(docker_client: DockerClient) -> "ImageManager":
        """ Creates an image from the ENVY_CONFIG environment

        Arguments:
            docker_client {DockerClient} -- A docker client

        Returns:
            [ImageManager] -- an image manager for the created image
        """
        print("Creating ENVy environment image.")

        # TODO: Use the correct image creator based on the config file's base image
        image_id = AptImageCreator(docker_client).create_image(
            ENVY_CONFIG.get_system_packages()
        )

        return ImageManager(docker_client, image_id)

    ### Image Management ###
    def __init__(self, docker_client: DockerClient, image_id: str):
        """ Creates an image manager for the given image id

        Arguments:
            docker_client {DockerClient} -- A docker client
            image_id {str} -- the image ID

        Returns:
            ImageManager -- an image manager for the created image
        """
        self.docker_client = docker_client

        self.image_id = image_id

    def __find(self) -> Image:
        for image in self.docker_client.images.list():
            if image.id == self.image_id:
                return image
        return None

    def destroy(self):
        """ Destroys the image
        """
        if self.image_id and self.__find():
            self.docker_client.images.remove(self.image_id)



#!/usr/bin/python3

import argparse

from envy.lib.config import ENVY_CONFIG
from envy.lib.state import ENVY_STATE, create_directory_if_not_exists
from envy.lib.docker_manager import ComposeManager, DockerManager
from envy.lib.setup_step import Builder

STATUS_MSG_NO_CONTAINER = "ENVy has not been initialized for this project. Please run `envy up` to install the ENVy environment."
STATUS_MSG_CONTAINER_STOPPED = (
    "ENVy is not running. Run `envy up` to start the ENVy environment."
)
STATUS_MSG_CONTAINER_READY = "ENVy environment is ready!"


def up_command(_args: argparse.Namespace, _unknow_args: [str]):
    docker_manager = DockerManager()

    if not docker_manager.connection_ok():
        docker_manager.print_connection_err()
        return

    if (
        ENVY_STATE.get_image_hash() != ENVY_CONFIG.get_image_hash()
        or ENVY_STATE.get_container_hash() != ENVY_CONFIG.get_container_hash()
    ):
        if ENVY_STATE.get_image_id:
            print(
                "Detected change in config environment. Re-creating ENVy environment."
            )
            docker_manager.nuke()
            ENVY_STATE.nuke()
            create_directory_if_not_exists()

    container = docker_manager.ensure_container()
    container.ensure_running()

    step_builder = Builder(container)
    step_builder.build()

    ENVY_STATE.set_image_hash(ENVY_CONFIG.get_image_hash())
    ENVY_STATE.set_container_hash(ENVY_CONFIG.get_container_hash())
    print(STATUS_MSG_CONTAINER_READY)

    compose_path = ENVY_CONFIG.get_services_compose_path()
    if compose_path:
        ComposeManager(compose_path).up()
        print("Sidecar services started")


def shell_command(_args: argparse.Namespace, _unknown_args: [str]):
    docker_manager = DockerManager()

    if not docker_manager.connection_ok():
        docker_manager.print_connection_err()
        return

    container = docker_manager.get_container()

    if not container:
        print(STATUS_MSG_NO_CONTAINER)
        return

    if not container.is_running():
        print(STATUS_MSG_CONTAINER_STOPPED)
        return

    container.exec("/bin/bash")


def down_command(_args: argparse.Namespace, _unknown_args: [str]):
    docker_manager = DockerManager()

    if not docker_manager.connection_ok():
        docker_manager.print_connection_err()
        return

    container = docker_manager.get_container()
    if container:
        container.ensure_stopped()

    print("ENVy environment stopped")

    compose_path = ENVY_CONFIG.get_services_compose_path()
    if compose_path:
        ComposeManager(compose_path).down()
        print("Sidecar services stopped")


def nuke_command(_args: argparse.Namespace, _unknown_args: [str]):
    docker_manager = DockerManager()

    if not docker_manager.connection_ok():
        docker_manager.print_connection_err()
        return

    docker_manager.nuke()
    ENVY_STATE.nuke()

    compose_path = ENVY_CONFIG.get_services_compose_path()
    if compose_path:
        ComposeManager(compose_path).nuke()

    print("ENVy environment destroyed")


def status_command(_args: argparse.Namespace, _unknown_args: [str]):
    docker_manager = DockerManager()

    if not docker_manager.connection_ok():
        docker_manager.print_connection_err()
        return

    container = docker_manager.get_container()

    if not container:
        print(STATUS_MSG_NO_CONTAINER)
    elif not container.is_running():
        print(STATUS_MSG_CONTAINER_STOPPED)
    else:
        print("ENVy environment is running!")


def run_script(_args: argparse.Namespace, unknown_args: [str], script: str):
    docker_manager = DockerManager()

    if not docker_manager.connection_ok():
        docker_manager.print_connection_err()
        return

    container = docker_manager.get_container()

    if not container:
        print(STATUS_MSG_NO_CONTAINER)
    elif not container.is_running():
        print(STATUS_MSG_CONTAINER_STOPPED)
    else:
        command = "{} {}".format(script, " ".join(unknown_args))
        container.exec(command)


def build_custom_command_parser(subparsers, name: str, info: {}):
    parser_custom = subparsers.add_parser(name, help=info.get("help"), add_help=False)
    parser_custom.set_defaults(
        func=lambda args, unknown_args: run_script(args, unknown_args, info["script"])
    )


def get_parser(actions: [{}]) -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Use ENVy to set up and use your development environment"
    )
    subparsers = parser.add_subparsers(dest="subparser_name")
    # Create 'up' parser
    parser_up = subparsers.add_parser(
        "up", help="Create and start the development environment"
    )
    parser_up.set_defaults(func=up_command)

    # Create 'shell' parser
    parser_shell = subparsers.add_parser(
        "shell", help="Enter a shell in the development environment"
    )
    parser_shell.set_defaults(func=shell_command)

    # Create 'down' parser
    parser_down = subparsers.add_parser("down", help="Stop the development environment")
    parser_down.set_defaults(func=down_command)

    # Create 'nuke' parser
    parser_nuke = subparsers.add_parser(
        "nuke", help="Destroy all ENVy data for this project"
    )
    parser_nuke.set_defaults(func=nuke_command)

    # Create 'status' parser
    parser_status = subparsers.add_parser("status", help="Show ENVy status")
    parser_status.set_defaults(func=status_command)

    # Create parsers for arbitrary custom commands
    for action in actions:
        build_custom_command_parser(subparsers, action["name"], action)

    return parser


def main():
    parser = get_parser(ENVY_CONFIG.get_actions())
    args, unknown = parser.parse_known_args()
    if args.subparser_name:
        args.func(args, unknown)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



from envy.lib.docker_manager import ContainerManager

from .assignable_trigger_step import AssignableTriggerStep


class ScriptSetupStep(AssignableTriggerStep):
    def __init__(self, name: str, container: ContainerManager, scripts: [str]):
        super().__init__(name, container)
        self._scripts = scripts

    def run(self):
        super().run()

        if isinstance(self._scripts, str):
            self._container.exec(self._scripts)
        else:
            for step in self._scripts:
                self._container.exec(step)



from .package_manager_step import PackageManagerStep


class AptPackageManagerStep(PackageManagerStep):
    """ An apt implementation for a Package Manager Step
    """

    def updated_packages(self) -> [str]:
        # TODO: consider changing the implementation to be aware of partial updates
        if self.has_built():
            return self._packages
        return []

    def should_trigger(self) -> bool:
        # TODO: this needs to be implemented before this class is used
        raise NotImplementedError

    def persist_trigger(self):
        # TODO: this needs to be implemented before this class is used
        raise NotImplementedError

    def run(self):
        super().run()
        versioned_packages = [
            "{}={}".format(package["recipe"], package["version"])
            if package["version"]
            else package["recipe"]
            for package in self._packages
        ]

        apt_string = " ".join(
            [versioned_package["recipe"] for versioned_package in versioned_packages]
        )
        self._container.exec(
            "apt-get update && apt-get install -y {} && rm -rf /var/lib/apt/lists/*".format(
                apt_string
            )
        )
